/* valaenum.vala
 *
 * Copyright (C) 2006  Jürg Billeter
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
 *
 * Author:
 * 	Jürg Billeter <j@bitron.ch>
 */
#define VALA_FREE_CHECKED(o,f) ((o) == NULL ? NULL : ((o) = (f (o), NULL)))
#define VALA_FREE_UNCHECKED(o,f) ((o) = (f (o), NULL))

#include "valaenum.h"
#include <vala/valasourcereference.h>
#include <vala/valaenumvalue.h>
#include <vala/valaenum.h>
#include <vala/valacodevisitor.h>
#include <vala/valanamespace.h>
#include <vala/valaattribute.h>
#include <vala/valanamedargument.h>
#include <vala/valaliteralexpression.h>
#include <vala/valaliteral.h>
#include <vala/valastringliteral.h>

struct _ValaEnumPrivate {
	GList* values;
	char* cname;
	char* cprefix;
};
#define VALA_ENUM_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), VALA_TYPE_ENUM, ValaEnumPrivate))
enum  {
	VALA_ENUM_DUMMY_PROPERTY,
};
static void vala_enum_real_accept (ValaCodeNode* base, ValaCodeVisitor* visitor);
static char* vala_enum_real_get_cname (ValaDataType* base, gboolean const_type);
static char* vala_enum_real_get_upper_case_cname (ValaDataType* base, char* infix);
static gboolean vala_enum_real_is_reference_type (ValaDataType* base);
static void vala_enum_set_cname (ValaEnum* self, char* cname);
static void vala_enum_process_ccode_attribute (ValaEnum* self, ValaAttribute* a);
static char* vala_enum_real_get_type_id (ValaDataType* base);
static char* vala_enum_real_get_marshaller_type_name (ValaDataType* base);
static char* vala_enum_real_get_get_value_function (ValaDataType* base);
static char* vala_enum_real_get_set_value_function (ValaDataType* base);
static gpointer vala_enum_parent_class = NULL;
static void vala_enum_dispose (GObject * obj);


/**
 * Creates a new enum.
 *
 * @param name   type name
 * @param source reference to source code
 * @return       newly created enum
 */
ValaEnum* vala_enum_new (char* _name, ValaSourceReference* source)
{
	g_return_val_if_fail (_name != NULL, NULL);
	g_return_val_if_fail (source == NULL || VALA_IS_SOURCE_REFERENCE (source), NULL);
	GParameter * __params = g_new0 (GParameter, 2);
	GParameter * __params_it = __params;
	(__params_it->name = "name", g_value_init (&__params_it->value, G_TYPE_STRING), g_value_set_string (&__params_it->value, _name), __params_it++);
	(__params_it->name = "source-reference", g_value_init (&__params_it->value, VALA_TYPE_SOURCE_REFERENCE), g_value_set_object (&__params_it->value, source), __params_it++);
	ValaEnum * self = g_object_newv (VALA_TYPE_ENUM, __params_it - __params, __params);
	return self;
}


/**
 * Appends the specified enum value to the list of values.
 *
 * @param value an enum value
 */
void vala_enum_add_value (ValaEnum* self, ValaEnumValue* value)
{
	g_return_if_fail (VALA_IS_ENUM (self));
	g_return_if_fail (VALA_IS_ENUM_VALUE (value));
	self->priv->values = g_list_append (self->priv->values, g_object_ref (value));
}


static void vala_enum_real_accept (ValaCodeNode* base, ValaCodeVisitor* visitor)
{
	ValaEnum * self = VALA_ENUM (base);
	g_return_if_fail (VALA_IS_CODE_VISITOR (visitor));
	vala_code_visitor_visit_begin_enum (visitor, self);
	{
		GList* __temp0 = NULL;
		__temp0 = self->priv->values;
		GList* value_it;
		for (value_it = __temp0; value_it != NULL; value_it = value_it->next) {
			ValaEnumValue* value = value_it->data;
			{
				vala_code_node_accept (VALA_CODE_NODE (value), visitor);
			}
		}
	}
	vala_code_visitor_visit_end_enum (visitor, self);
}


static char* vala_enum_real_get_cname (ValaDataType* base, gboolean const_type)
{
	ValaEnum * self = VALA_ENUM (base);
	if (self->priv->cname == NULL) {
		char* __temp1 = NULL;
		self->priv->cname = (__temp1 = g_strdup_printf ("%s%s", vala_namespace_get_cprefix (VALA_DATA_TYPE (self)->namespace), vala_data_type_get_name (VALA_DATA_TYPE (self))), (self->priv->cname == NULL ? NULL : (self->priv->cname = (g_free (self->priv->cname), NULL))), __temp1);
	}
	return self->priv->cname;
}


static char* vala_enum_real_get_upper_case_cname (ValaDataType* base, char* infix)
{
	ValaEnum * self = VALA_ENUM (base);
	char* __temp4 = NULL;
	char* __temp3 = NULL;
	char* __temp5 = NULL;
	return (__temp5 = g_utf8_strup ((__temp4 = g_strdup_printf ("%s%s", vala_namespace_get_lower_case_cprefix (VALA_DATA_TYPE (self)->namespace), (__temp3 = vala_namespace_camel_case_to_lower_case (vala_data_type_get_name (VALA_DATA_TYPE (self)))))), -1), (__temp4 == NULL ? NULL : (__temp4 = (g_free (__temp4), NULL))), (__temp3 == NULL ? NULL : (__temp3 = (g_free (__temp3), NULL))), __temp5);
}


static gboolean vala_enum_real_is_reference_type (ValaDataType* base)
{
	ValaEnum * self = VALA_ENUM (base);
	return FALSE;
}


static void vala_enum_set_cname (ValaEnum* self, char* cname)
{
	g_return_if_fail (VALA_IS_ENUM (self));
	g_return_if_fail (cname != NULL);
	char* __temp8 = NULL;
	self->priv->cname = (__temp8 = g_strdup (cname), (self->priv->cname == NULL ? NULL : (self->priv->cname = (g_free (self->priv->cname), NULL))), __temp8);
}


/**
 * Returns the string to be prepended to the name of members of this
 * enum when used in C code.
 *
 * @return the prefix to be used in C code
 */
char* vala_enum_get_cprefix (ValaEnum* self)
{
	g_return_val_if_fail (VALA_IS_ENUM (self), NULL);
	if (self->priv->cprefix == NULL) {
		char* __temp10 = NULL;
		char* __temp9 = NULL;
		self->priv->cprefix = (__temp10 = g_strdup_printf ("%s_", (__temp9 = vala_data_type_get_upper_case_cname (VALA_DATA_TYPE (self), NULL))), (self->priv->cprefix == NULL ? NULL : (self->priv->cprefix = (g_free (self->priv->cprefix), NULL))), __temp10);
		(__temp9 == NULL ? NULL : (__temp9 = (g_free (__temp9), NULL)));
	}
	return self->priv->cprefix;
}


/**
 * Sets the string to be prepended to the name of members of this enum
 * when used in C code.
 *
 * @param cprefix the prefix to be used in C code
 */
void vala_enum_set_cprefix (ValaEnum* self, char* cprefix)
{
	g_return_if_fail (VALA_IS_ENUM (self));
	g_return_if_fail (cprefix != NULL);
	char* __temp12 = NULL;
	self->priv->cprefix = (__temp12 = g_strdup (cprefix), (self->priv->cprefix == NULL ? NULL : (self->priv->cprefix = (g_free (self->priv->cprefix), NULL))), __temp12);
}


static void vala_enum_process_ccode_attribute (ValaEnum* self, ValaAttribute* a)
{
	g_return_if_fail (VALA_IS_ENUM (self));
	g_return_if_fail (VALA_IS_ATTRIBUTE (a));
	{
		GList* __temp16 = NULL;
		__temp16 = a->args;
		GList* arg_it;
		for (arg_it = __temp16; arg_it != NULL; arg_it = arg_it->next) {
			ValaNamedArgument* arg = arg_it->data;
			{
				if (g_utf8_collate (vala_named_argument_get_name (arg), "cname") == 0) {
					/* this will already be checked during semantic analysis */
					if (VALA_IS_LITERAL_EXPRESSION (vala_named_argument_get_argument (arg))) {
						ValaLiteral* lit = vala_literal_expression_get_literal ((VALA_LITERAL_EXPRESSION (vala_named_argument_get_argument (arg))));
						if (VALA_IS_STRING_LITERAL (lit)) {
							char* __temp13 = NULL;
							vala_enum_set_cname (self, (__temp13 = vala_string_literal_eval ((VALA_STRING_LITERAL (lit)))));
							(__temp13 == NULL ? NULL : (__temp13 = (g_free (__temp13), NULL)));
						}
					}
				} else {
					if (g_utf8_collate (vala_named_argument_get_name (arg), "cprefix") == 0) {
						/* this will already be checked during semantic analysis */
						if (VALA_IS_LITERAL_EXPRESSION (vala_named_argument_get_argument (arg))) {
							ValaLiteral* lit = vala_literal_expression_get_literal ((VALA_LITERAL_EXPRESSION (vala_named_argument_get_argument (arg))));
							if (VALA_IS_STRING_LITERAL (lit)) {
								char* __temp14 = NULL;
								vala_enum_set_cprefix (self, (__temp14 = vala_string_literal_eval ((VALA_STRING_LITERAL (lit)))));
								(__temp14 == NULL ? NULL : (__temp14 = (g_free (__temp14), NULL)));
							}
						}
					} else {
						if (g_utf8_collate (vala_named_argument_get_name (arg), "cheader_filename") == 0) {
							/* this will already be checked during semantic analysis */
							if (VALA_IS_LITERAL_EXPRESSION (vala_named_argument_get_argument (arg))) {
								ValaLiteral* lit = vala_literal_expression_get_literal ((VALA_LITERAL_EXPRESSION (vala_named_argument_get_argument (arg))));
								if (VALA_IS_STRING_LITERAL (lit)) {
									char* val = vala_string_literal_eval ((VALA_STRING_LITERAL (lit)));
									{
										char** __temp15 = NULL;
										__temp15 = g_strsplit (val, ",", 0);
										char** filename_it;
										for (filename_it = __temp15; *filename_it != NULL; filename_it = filename_it + 1) {
											char* filename = *filename_it;
											{
												vala_data_type_add_cheader_filename (VALA_DATA_TYPE (self), filename);
											}
										}
										(__temp15 == NULL ? NULL : (__temp15 = (g_strfreev (__temp15), NULL)));
									}
									(val == NULL ? NULL : (val = (g_free (val), NULL)));
								}
							}
						}
					}
				}
			}
		}
	}
}


/**
 * Process all associated attributes.
 */
void vala_enum_process_attributes (ValaEnum* self)
{
	g_return_if_fail (VALA_IS_ENUM (self));
	{
		GList* __temp17 = NULL;
		__temp17 = VALA_CODE_NODE (self)->attributes;
		GList* a_it;
		for (a_it = __temp17; a_it != NULL; a_it = a_it->next) {
			ValaAttribute* a = a_it->data;
			{
				if (g_utf8_collate (vala_attribute_get_name (a), "CCode") == 0) {
					vala_enum_process_ccode_attribute (self, a);
				}
			}
		}
	}
}


static char* vala_enum_real_get_type_id (ValaDataType* base)
{
	ValaEnum * self = VALA_ENUM (base);
	return "G_TYPE_INT";
}


/* FIXME: use GType-registered enums*/
static char* vala_enum_real_get_marshaller_type_name (ValaDataType* base)
{
	ValaEnum * self = VALA_ENUM (base);
	return "ENUM";
}


static char* vala_enum_real_get_get_value_function (ValaDataType* base)
{
	ValaEnum * self = VALA_ENUM (base);
	return "g_value_get_enum";
}


static char* vala_enum_real_get_set_value_function (ValaDataType* base)
{
	ValaEnum * self = VALA_ENUM (base);
	return "g_value_set_enum";
}


static void vala_enum_get_property (GObject * object, guint property_id, GValue * value, GParamSpec * pspec)
{
	ValaEnum * self = VALA_ENUM (object);
	switch (property_id) {
	}
}


static void vala_enum_set_property (GObject * object, guint property_id, const GValue * value, GParamSpec * pspec)
{
	ValaEnum * self = VALA_ENUM (object);
	switch (property_id) {
	}
}


static void vala_enum_class_init (ValaEnumClass * klass)
{
	vala_enum_parent_class = g_type_class_peek_parent (klass);
	g_type_class_add_private (klass, sizeof (ValaEnumPrivate));
	G_OBJECT_CLASS (klass)->get_property = vala_enum_get_property;
	G_OBJECT_CLASS (klass)->set_property = vala_enum_set_property;
	G_OBJECT_CLASS (klass)->dispose = vala_enum_dispose;
	VALA_CODE_NODE_CLASS (klass)->accept = vala_enum_real_accept;
	VALA_DATA_TYPE_CLASS (klass)->get_cname = vala_enum_real_get_cname;
	VALA_DATA_TYPE_CLASS (klass)->get_upper_case_cname = vala_enum_real_get_upper_case_cname;
	VALA_DATA_TYPE_CLASS (klass)->is_reference_type = vala_enum_real_is_reference_type;
	VALA_DATA_TYPE_CLASS (klass)->get_type_id = vala_enum_real_get_type_id;
	VALA_DATA_TYPE_CLASS (klass)->get_marshaller_type_name = vala_enum_real_get_marshaller_type_name;
	VALA_DATA_TYPE_CLASS (klass)->get_get_value_function = vala_enum_real_get_get_value_function;
	VALA_DATA_TYPE_CLASS (klass)->get_set_value_function = vala_enum_real_get_set_value_function;
}


static void vala_enum_init (ValaEnum * self)
{
	self->priv = VALA_ENUM_GET_PRIVATE (self);
}


static void vala_enum_dispose (GObject * obj)
{
	ValaEnum * self = VALA_ENUM (obj);
	(self->priv->values == NULL ? NULL : (self->priv->values = (g_list_foreach (self->priv->values, (GFunc) g_object_unref, NULL), g_list_free (self->priv->values), NULL)));
	(self->priv->cname == NULL ? NULL : (self->priv->cname = (g_free (self->priv->cname), NULL)));
	(self->priv->cprefix == NULL ? NULL : (self->priv->cprefix = (g_free (self->priv->cprefix), NULL)));
	ValaEnumClass * klass;
	GObjectClass * parent_class;
	klass = VALA_ENUM_CLASS (g_type_class_peek (VALA_TYPE_ENUM));
	parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
	parent_class->dispose (obj);
}


GType vala_enum_get_type ()
{
	static GType g_define_type_id = 0;
	if (G_UNLIKELY (g_define_type_id == 0)) {
		static const GTypeInfo g_define_type_info = { sizeof (ValaEnumClass), (GBaseInitFunc) NULL, (GBaseFinalizeFunc) NULL, (GClassInitFunc) vala_enum_class_init, (GClassFinalizeFunc) NULL, NULL, sizeof (ValaEnum), 0, (GInstanceInitFunc) vala_enum_init };
		g_define_type_id = g_type_register_static (VALA_TYPE_DATA_TYPE, "ValaEnum", &g_define_type_info, 0);
	}
	return g_define_type_id;
}




